/*
   core/gameinterface.h
   This file is part of the Osirion project and is distributed under
   the terms of the GNU General Public License version 2
*/

#ifndef __INCLUDED_CORE_GAMEINTERFACE_H__
#define __INCLUDED_CORE_GAMEINTERFACE_H__

#include "core/player.h"
#include "core/info.h"
#include "model/vertexarray.h"

namespace core
{

/// abstract interface from the core to the game-specific code
class GameInterface
{
public:
	/// create a new game
	GameInterface();

	/// destroy the game
	virtual ~GameInterface();

	/// type definition for the Players collection
	typedef std::list<Player *> Players;

/*----- inspectors ---------------------------------------------- */

	/// return the local player
	inline Player *localplayer() { return &game_localplayer; }

	inline Players & players() { return game_players; }

	inline model::VertexArray *vertexarray() { return game_vertexarray; }
	
	/// find the first player who's id or name matches the search string
	Player *find_player(const std::string &search);
	
	/// show a list of connected players
	void list_players();

/*----- virtual inspectors --------------------------------------- */

	/// returns true if the game server can run a time frime
	virtual bool running() const = 0;

	/// returns true if the game is running an interactive module
	virtual bool interactive() const = 0;

	/// return the current game time in seconds
	virtual float time() const = 0;

	/// return the current game time
	virtual unsigned long timestamp() const = 0;

	/// returns an info record
	virtual Info *info(const std::string &label) = 0;

/*-----  mutators ------------------------------------------------- */
	
	/// clear all game variables, game functions and entities
	void clear();


/*----- virtual mutators ------------------------------------------ */
	
	/// run one game time frame
	/// @param timestamp current application time
	virtual void frame(unsigned long timestamp) = 0;

protected:
	/// the local player
	static Player 			game_localplayer;

	/// all the players
	Players				game_players;

	model::VertexArray 		*game_vertexarray;
};

/// global local player instance
Player *localplayer();

/// global local control instance
EntityControlable *localcontrol();

}

#endif // __INCLUDED_CORE_GAMEINTERFACE_H__
